#include "Drawables.h"

namespace GWCTutorials{

	TouchPoint::TouchPoint()
		: id(-1)
		, x(0)
		, y(0)
		, radius(30)
		, color(0)
	{}

	TouchPoint::TouchPoint(long id, int x, int y)
		: id(id)
		, x(x)
		, y(y)
		, radius(30)
		, color(0xff0000)
	{}

	void TouchPoint::SetPosition(int new_x, int new_y){
		x = new_x;
		y = new_y;
		x = min(max(0, x), ofGetWindowWidth());
		y = min(max(0, y), ofGetWindowHeight());
	}

	void TouchPoint::SetColor(unsigned int new_color){
		color = new_color;
	}

	void TouchPoint::SetRadius(int new_radius){
		radius = new_radius;
	}

	void TouchPoint::Draw() const{
		ofSetHexColor(color);
		ofFill();
		ofCircle(x, y, radius - 10);
		ofNoFill();
		ofSetLineWidth(2);
		ofCircle(x, y, radius);

		std::stringstream info;
		info << "ID: " << id << " X: " << x << " Y: " << y;
		ofDrawBitmapString(info.str(), x + 40, y - 40);
	}

	TouchRectangle::TouchRectangle()		
		: width(100)
		, rotation(0.f)
		, scale(1.f)
		, x(0)
		, y(0)
	{}

	TouchRectangle::TouchRectangle(const ofImage& image, int width)
		: image(image)		
		, width(width)
		, rotation(0.f)
		, scale(1.f)
		, x(0)
		, y(0)
	{}

	void TouchRectangle::SetPosition(int new_x, int new_y){
		x = new_x;
		y = new_y;
		x = min(max(0, x), ofGetWindowWidth());
		y = min(max(0, y), ofGetWindowHeight());
	}

	void TouchRectangle::IncrementPosition(int dx, int dy){		
		SetPosition(x + dx, y + dy);
	}

	void TouchRectangle::SetScale(float new_scale){
		scale = new_scale;
		scale = min(max(.5f, scale), 3.f);
	}

	void TouchRectangle::IncrementScale(float ds){
		SetScale(scale + ds);
	}

	void TouchRectangle::Rotate(float dt, float cx, float cy){
		rotation += dt;
		if (!owned_points.size()){
			// If the object is being rotated by inertia only, just rotate about the center
			return;
		}

		// Rotate the object about the center of the interaction point
		float rads = PI * dt / 180.f;
		std::pair<int, int> new_pos = rotateAboutPoint(x, y, cx, cy, rads);
		SetPosition(new_pos.first, new_pos.second);
	}

	void TouchRectangle::Draw(){	
		ofSetHexColor(0xffffff);
		ofSetRectMode(OF_RECTMODE_CENTER);		
		ofPushMatrix();
		ofTranslate(x, y);
		ofRotateZ(rotation);
		ofScale(scale, scale, 1);
		image.draw(0, 0, width, width);
		ofPopMatrix();
	}

	// Returns true if (px, py) intersects our object
	bool TouchRectangle::Hit(int px, int py) const{
		float rads = PI * rotation / 180.f;
		std::pair<int, int> local_coords(rotateAboutPoint(px, py, x, y, rads));

		if (abs(local_coords.first - x) <= width * scale / 2){
			if (abs(local_coords.second - y) <= width * scale / 2){
				return true;
			}
		}
		return false;
	}

	// Keep track of the points that "belong" to this object
	bool TouchRectangle::AddPoint(long id){
		auto search = std::find(owned_points.begin(), owned_points.end(), id);
		if (search != owned_points.end()){
			return false;
		}

		owned_points.push_back(id);
		return true;
	}

	bool TouchRectangle::RemovePoint(long id){
		auto search = std::find(owned_points.begin(), owned_points.end(), id);
		if (search == owned_points.end()){
			return false;
		}

		owned_points.erase(search);
		return true;
	}

	std::pair<int, int> normScreenToWindowPx(float screen_x, float screen_y){
		ofRectangle rect = ofGetWindowRect();
		int x_px = static_cast<int>(screen_x * ofGetScreenWidth());
		int y_px = static_cast<int>(screen_y * ofGetScreenHeight());
		int x = x_px - ofGetWindowPositionX();
		int y = y_px - ofGetWindowPositionY();
		return std::pair<int, int>(x, y);
	}

	std::pair<int, int> rotateAboutPoint(
		int point_x,
		int point_y,
		int rotation_center_x,
		int rotation_center_y,
		float angle_in_rads)
	{
		float dx = static_cast<float>(point_x - rotation_center_x);
		float dy = static_cast<float>(point_y - rotation_center_y);
		float len = sqrt(dx * dx + dy * dy);

		float arctan = atan2(dy, dx);

		float new_x = len * cos(angle_in_rads + arctan) + static_cast<float>(rotation_center_x);
		float new_y = len * sin(angle_in_rads + arctan) + static_cast<float>(rotation_center_y);

		return std::pair<int, int>(static_cast<int>(new_x), static_cast<int>(new_y));
	}
}