#include "InteractiveBitmaps.h"

using namespace GWCTutorials;

const std::string InteractiveBitmaps::window_title("Interactive Bitmaps");

InteractiveBitmaps::InteractiveBitmaps()
: last_tick(0)
{}

// openFrameworks calls this function once before the app starts
void InteractiveBitmaps::setup(){
	if (!GestureWorks::Instance()->LoadGestureWorks(L"GestureWorksCore32.dll")){
		MessageBox(NULL, L"Unable to load GestureWorksCore DLL", L"Error", 0);
		return;
	}

	// Initialize with (0,0) so GestureWorks will automatically detect our screen resolution
	GestureWorks::Instance()->InitializeGestureWorks(0, 0);

	// Register our window with GestureWorks for touch input. We could also do this with an HWND.
	// Alternatively, we could use some other method of getting touch input and feed it to GestureWorks via
	// the addEvent method.
	if (!GestureWorks::Instance()->RegisterWindowForTouchByName(std::wstring(window_title.begin(), window_title.end()))){
		MessageBox(NULL, L"Unable to register window for touch", L"Error", 0);
		return;
	}

	if (!GestureWorks::Instance()->LoadGML(L"basic_manipulation.gml")){					
		MessageBox(NULL, L"Unable to load GML", L"Error", 0);
		return;		
	}

	logo.loadImage("gw_logo.png");

	TouchRectangle rect(logo, 200);
	int center_x = ofGetWindowWidth() / 2;
	int center_y = ofGetWindowHeight() / 2;
	rect.SetPosition(center_x - 200, center_y);
	rectangles[L"logo1"] = rect;

	rect.SetPosition(center_x + 200, center_y);
	rectangles[L"logo2"] = rect;

	// Tell GestureWorks about our touch objects
	GestureWorks::Instance()->RegisterTouchObject(L"logo1");
	GestureWorks::Instance()->RegisterTouchObject(L"logo2");

	// Now we'll add gesture that are defined in the GML file we loaded earlier

	// Add an entire gesture set to "logo1"
	GestureWorks::Instance()->AddGestureSet(L"logo1", L"basic-gestures");

	// We could also add gestures individually like we do for "logo2"
	GestureWorks::Instance()->AddGesture(L"logo2", L"n-drag");
	GestureWorks::Instance()->AddGesture(L"logo2", L"n-scale");
	GestureWorks::Instance()->AddGesture(L"logo2", L"n-rotate");
}

// openFrameworks calls this function before every draw cycle
void InteractiveBitmaps::update(){
	// For consistency, we try to run GestureWorks at a constant rate - about 60Hz in this case
	clock_t this_tick(clock());
	if (this_tick - last_tick < 16) return;
	last_tick = this_tick;

	// Tell GestureWorks to process all of the data since the last frame
	GestureWorks::Instance()->ProcessFrame();

	// Get all the point events from the current frame
	std::vector<gwc::PointEvent> point_events = GestureWorks::Instance()->ConsumePointEvents();
	handlePointEvents(point_events);

	// Get all the gesture events from the current frame
	std::vector<gwc::GestureEvent> gesture_events = GestureWorks::Instance()->ConsumeGestureEvents();
	handleGestureEvents(gesture_events);
}

// openFrameworks calls this function in an infinite loop until the app ends 
void InteractiveBitmaps::draw(){
	for (auto rect : rectangles){
		rect.second.Draw();
	}

	for (const auto& point : touch_points){
		point.second.Draw();
	}
}

void InteractiveBitmaps::handlePointEvents(const std::vector<gwc::PointEvent>& point_events){
	for (const auto& point : point_events){
		// By default, GestureWorks sends us coordinates in normalized screen space. To get to localized pixel
		// coordinates we need to do a conversion.
		std::pair<int, int> window_coords(normScreenToWindowPx(point.position.x, point.position.y));

		if (point.status == gwc::TOUCHADDED){
			for (auto& rect : rectangles){
				if (rect.second.Hit(window_coords.first, window_coords.second)){

					// If this is a new point and hits one of our touch objects, we want to let GestureWorks
					// know so that it can perform cluster analysis during its processing cycle. 
					GestureWorks::Instance()->AddTouchPoint(rect.first, point.point_id);				
					rect.second.AddPoint(point.point_id);
					break; // only get the first one
				}
			}
		}		
		else if (point.status == gwc::TOUCHUPDATE){			
			touch_points[point.point_id] = TouchPoint(point.point_id, window_coords.first, window_coords.second);
		}
		else{
			touch_points.erase(point.point_id);
			for (auto& rect : rectangles){
				rect.second.RemovePoint(point.point_id);
			}
		}
	}
}

void InteractiveBitmaps::handleGestureEvents(const std::vector<gwc::GestureEvent>& gesture_events){
	for (const auto& gesture : gesture_events){
		std::wstring target(gesture.target.begin(), gesture.target.end());

		// Find the name of the touch object the gesture "targets". This will be the same name we used to register
		// the touch object with GestureWorks
		auto rect = rectangles.find(target);
		if (rect == rectangles.end()) continue;

		if (gesture.gesture_id == "n-drag"){
			int dx = static_cast<int>(gesture.values.at("drag_dx") * ofGetScreenWidth());
			int dy = static_cast<int>(gesture.values.at("drag_dy") * ofGetScreenHeight());			
			rect->second.IncrementPosition(dx, dy);
		}
		else if (gesture.gesture_id == "n-scale"){
			float ds = gesture.values.at("scale_dsx") * ofGetScreenWidth();			
			rect->second.IncrementScale(ds);
		}
		else if (gesture.gesture_id == "n-rotate"){
			float dtheta = gesture.values.at("rotate_dtheta");
			if (!dtheta) continue;

			std::pair<int, int> center(normScreenToWindowPx(gesture.x, gesture.y));
			rect->second.Rotate(dtheta, center.first, center.second);
		}
	}
}
